/******************************************************************************
 * $Id: file.c,v 1.15 2000/09/09 22:03:12 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for files.
 ******************************************************************************/

#include "header.h"

/******************************************************************************
 * Load a regular file into a memory buffer.
 * For special files, instead use LoadFileSpecial().
 *
 * The buffer will be appended with a null char, but will not affect
 * the returned file size.
 *
 * The size of the file must be > 0.
 * The size of the file is not expected to change during this function.
 * If it does, this function will interpret a file size change as an error.
 *
 * Returns : Pointer to memory buffer allocated by ** g_malloc() **
 *           -or- NULL upon failure.
 *           The size of the buffer is 1 char greater than file size.
 *
 * Parms   : pFileName
 *
 *         : pFileSize (out)
 *           Excludes null char that will be appended to buffer.
 ******************************************************************************/
gpointer
LoadFile( gchar* pFileName,
          long*  pFileSize /* out */ )
{
   FILE*	pFile	= NULL;
   gpointer	pBuffer	= NULL;
   size_t	bytes	= 0;

  /*
   * Determine the size of the file.
   * Cause an error if the size is not > 0.
   */
   *pFileSize = FileSize( pFileName );
   if ( *pFileSize < 1 ) return NULL;

  /*
   * Allocate the memory buffer according to the size of the file.
   * Include an extra byte for a null char to append.
   */
   pBuffer = g_malloc( *pFileSize + 1 );

  /*
   * Open the file for reading.
   */
   pFile = fopen( pFileName, "r" );

  /*
   * Read the file into the memory buffer.
   */
   bytes = \
   fread( pBuffer,
          1,
          *pFileSize,
          pFile );

  /*
   * Close the file.
   */
   fclose( pFile );

  /*
   * Append a null char.
   */
   ((char*)pBuffer)[*pFileSize] = '\0';

  /*
   * Return a pointer to the buffer if successful.
   */
   if ( (bytes == *pFileSize) && bytes ) return pBuffer;

  /*
   * An error occurred, free the buffer, return NULL.
   */
   g_free( pBuffer );
   Popup( "Failed to load file", "Error", FONT_PROPORTIONAL );
   return NULL;
}

/******************************************************************************
 * Load a special file (whose size can't be predetermined) into a GString.
 *
 * Returns : Pointer to GString or NULL if failure.
 *           Note that zero-length is not considered a failure here.
 *
 * Parms   : pFileName
 ******************************************************************************/
GString*
LoadFileSpecial( gchar* pFileName )
{
   FILE*	pFile	= NULL;
   char		bufChar[1];
   GString* pGstr;

  /*
   * Create GString, open the file for reading.
   */
   pGstr = g_string_sized_new( 0 );
   if ( pGstr == NULL ) return NULL;

   pFile = fopen( pFileName, "r" );
   if ( pFile == NULL ) return NULL;

  /*
   * Read the file, one char at a time, into a GString, until EOF.
   */
   for (;;)
   {
      int cnt;

     /*
      * Read next char.
      */
      cnt = \
      fread( &bufChar,
             1,
             1,
             pFile );

     /*
      * If present, append next char into GString, else break.
      */
      if ( cnt )
         pGstr = g_string_append_c( pGstr, bufChar[0] );
      else
         break;
   }

  /*
   * Close the file.
   */
   fclose( pFile );

  /*
   * Return GString.
   */
   return pGstr;
}

/******************************************************************************
 * Return the size of a file (returns negative upon failure).
 ******************************************************************************/
long
FileSize( gchar* pFilename )
{
   FILE*	file	= NULL;
   long		fsize	= -1;

  /*
   * Try to open the file for reading.
   */
   file = fopen( pFilename, "r" );

  /*
   * Ensure the file was opened.
   */
   if ( file == NULL ) return FUNC_FAIL_NEGATIVE;

  /*
   * Use fseek() and ftell() to determine the file size.
   *
   * fseek() returns 0 upon success.
   * ftell() returns the current offset (ie the file size in this case).
   *
   */
   if ( fseek( file, 0, SEEK_END ) ) { fclose( file ); return FUNC_FAIL_NEGATIVE; }
   fsize = ftell( file );

  /*
   * Close the file and return its size.
   */
   fclose( file );
   return fsize;
}
