/******************************************************************************
 * $Id: find.c,v 1.46 2001/01/01 06:19:28 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for finding.
 ******************************************************************************/

#include "header.h"

/******************************************************************************
 * Find a string within a memory segment.
 *
 * Can search forwards or backwards, case-[in]senstive, depending on
 * "findReverse", "findCaseSensText".
 *
 * Parms   : pSought
 *           Pointer to the string to find.
 *
 *           pSegment
 *           Pointer to segment to search.
 *
 *           rangeLen
 *           Length of segment to search.
 *
 *           findCaseSensText (GLOBAL)
 *
 *           findReverse (GLOBAL)
 *
 * Returns : If found, pointer to the string, else NULL.
 ******************************************************************************/
gchar*
FindStringInSegment( gchar* pSought,
                     gchar* pSegment,
                     int    segmentLen )
{
   int	idx;
   int	soughtLen;

   g_return_val_if_fail( segmentLen > 0, NULL );
/* TODO use a sole #define for max text buffer */
   g_return_val_if_fail( segmentLen <= MAX_LINE_LEN * MAX_LINES, NULL );

   soughtLen = strlen( pSought );

   if ( soughtLen > segmentLen ) return NULL;

  /*
   * Search forward or reverse.
   */
   if ( ! findReverse )
   {
      /* TODO use FindSubstr() instead */
      for ( idx = 0; idx < segmentLen; ++idx )
      {
         if ( 0 == (findCaseSensText ? strncmp(     &pSegment[idx], pSought, soughtLen )
                                     : strncasecmp( &pSegment[idx], pSought, soughtLen )) )
         {
            return &pSegment[idx];
         }
      }
      return NULL;
      }
   else
   {
      return FindSubstrReverse( pSegment + segmentLen - 1,
                                pSegment,
                                pSought,
                                findCaseSensText ? strncmp : strncasecmp );
   }
}

/******************************************************************************
 * Find the next string in the text widget.
 *
 * First does a search starting from the current position.
 * If unsuccessful, tries a wrapped search.
 * If found, the matching string will be scrolled to and highlighted. 
 *
 * Parms   : pSoughtText (GLOBAL)
 *           Pointer to the previous sought string.
 *
 *           findCaseSensText (GLOBAL)
 *           Whether to be case-sensitive.
 *
 * Returns  : TRUE if found.
 ******************************************************************************/
gboolean
FindNextStringInTextWidget( void )
{
  /*
   * Has Find/Text been selected previously?
   */
   if ( soughtType == MENU_ITEM_FIND_TEXT  &&  pSoughtText )
   {
      return FindStringInTextWidget( pSoughtText->str,
                                     TRUE,		/* do wrap */
                                     FALSE,		/* don't be quiet */
                                     TRUE,      /* do scroll to match */
                                     NULL );    /* don't care about line # */
   }

  /*
   * Find/Text has not been selected yet.
   */
   return FALSE;
}

/*****************************************************************************
 * Find a tag in clist widget.
 * Warn user if there are multiple matches.
 * Returns row # of first match.
 *****************************************************************************/
int
FindTagCheckMultipleMatches( char*    pTagName,
                             int      rowStarting,
                             gboolean exact,
                             gboolean caseSens,
                             char*    pMsgMultipleMatches )
{
   gint		rowFirstMatch;
   gint		row;

   g_assert ( ! busy_PopulateTagsClistWidget );

   RecordCursorInTagHistory( "(recorded position)" );

  /*
   * Try to find first match.  If found, then select row.
   */
   rowFirstMatch = \
   SearchClistWidget( pTagName,
                      pWidgetClistTags,
                      rowStarting,
                      exact,
                      caseSens,
                      TRUE, /* be quiet */
                      TRUE,	/* OK to update notebook */
                      TRUE	/* do select row */ );

   if ( rowFirstMatch == ROW_NONE )
   {
      UndoRecordCursorInTagHistory();

      if ( exact )
         PrintStatusbarThirds( "Tag '",    pTagName, "' not found (exact search)." );
      else
         if ( caseSens )
            PrintStatusbarThirds( "Tag '",    pTagName, "' not found (fuzzy/case search)." );
         else
            PrintStatusbarThirds( "Tag '",    pTagName, "' not found (fuzzy search)." );

      return ROW_NONE;
   }

  /*
   * Check if there are multiple matches.
   * Print a message if so.
   */
   row = \
   SearchClistWidget( pTagName,
                      pWidgetClistTags,
                      widgetTagsClistSelectedRow,
                      exact,
                      caseSens,
                      TRUE,	/* quiet */
                      FALSE,/* don't update notebook */
                      FALSE	/* don't select row */ );

   if ( row != ROW_NONE  &&  row != widgetTagsClistSelectedRow )
      PrintStatusbar( pMsgMultipleMatches );

   return rowFirstMatch;
}

/*****************************************************************************
 * Repeat the previous Find/Text or Find/Tag.
 *****************************************************************************/
void
FindNextTextOrTag( void )
{
  /*
   * Repeat Find/Text ?
   */
   if ( soughtType == MENU_ITEM_FIND_TEXT )
   {
      FindNextStringInTextWidget();
      return;
   }

  /*
   * Repeat Find/Tag ?
   */
   if ( soughtType == MENU_ITEM_FIND_TAG )
   {
      FindNextTag();
      return;
   }

   PrintStatusbar( "No previous search." );
   return;
}

/*****************************************************************************
 * Repeat the last Find/Tag.
 *
 * Parms   : pSoughtTag (GLOBAL)
 *           Pointer to the name of sought tag.
 *
 *           [other find state] (GLOBAL)
 *
 * Returns : If found, the row in which the match was found (row will be >=0),
 *           else returns ROW_NONE.
 *****************************************************************************/
gint
FindNextTag( void )
{
   g_assert ( ! busy_PopulateTagsClistWidget );

  /*
   * Has Find/Tag been selected previously?
   */
   if ( soughtType == MENU_ITEM_FIND_TAG  &&  pSoughtTag->str )
   {
      return \
      SearchClistWidget( pSoughtTag->str,
                         pWidgetClistTags,
                         widgetTagsClistSelectedRow,
                         findExactTag,
                         findCaseSensTag,
                         FALSE,	/* be loud */
                         TRUE,	/* OK to update notebook */
                         TRUE	/* do select row */ );
   }

  /*
   * Find/Tag has not been selected yet.
   */
   return ROW_NONE;
}

/******************************************************************************
 * Find a string in the text widget.
 *
 * First does a search starting from the current position.
 * If unsuccessful, can try a wrapped search (if requested).
 * If found (and requested), the matching string will be scrolled to and highlighted. 
 *
 * Parms   : pSought
 *           Pointer to the string to find.
 *
 *           tryWrap
 *           If true then wrap search to top if necessary.
 *
 *           quiet
 *           If true then don't print to statusbar.
 *
 *           scroll
 *           If true then match will be scrolled to.
 *
 *           pReturnLineNumFound.
 *           Pass NULL if not wanted.
 *           Ignore this if FALSE is returned by this function.
 *
 *           findCaseSensText (GLOBAL)
 *
 *           findReverse (GLOBAL)
 *
 * Returns  : TRUE if found.
 ******************************************************************************/
gboolean
FindStringInTextWidget( gchar*   pSought,
                        gboolean tryWrap,
                        gboolean quiet,
                        gboolean scroll,
                        gint*    pReturnLineNumFound )
{
   gboolean found;
   gchar*	pText;
   gint		textLen;

  /*
   * Return if Find/Text (local) was selected but the text widget is empty.
   * A similar check for Find/Tag is unnecessary because there
   * should always be a list of tags.
   */
   if ( soughtType == MENU_ITEM_FIND_TEXT  &&  IsTextWidgetEmpty() )
   {
      PrintStatusbar( "Text widget is empty (try selecting 'global' checkbox)." );
      return FALSE;
   }

   if ( ! quiet ) PrintStatusbarThirdsNoTimeout( "Searching for string '", pSought, "'..." );

  /*
   * Compute pointer/length of the buffer of editable chars in the text widget.
   * Pointer to buffer will be free later, before returning.
   */
   gtk_text_freeze( GTK_TEXT(pWidgetText) );
   textLen = gtk_text_get_length( GTK_TEXT(pWidgetText) );
   pText = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   0,
                                   -1 );
   gtk_text_thaw( GTK_TEXT(pWidgetText) );

  /*
   * Do the actual search.
   */
   if ( pText  &&  textLen > 0 )
      found = FindStringInTextWidget_( pSought,
                                       tryWrap,
                                       quiet,
                                       scroll,
                                       pReturnLineNumFound,
                                       pText,
                                       textLen );
   else
      found = FALSE;

  /*
   * Free the text buffer, return whether match was found.
   */
   g_free( pText );
   return found;
}

/******************************************************************************
 * Primitive of FindStringInTextWidget().
 * FindStringInTextWidget() is responsible for managing the text memory buffer
 * that this primitive searches thru.
 *
 * (Separation was needed, because the old FindStringInTextWidget()
 *  wasn't matching gtk_editable_get_chars() and g_free() every time,
 *  resulting in core dumps >:-(
 ******************************************************************************/
gboolean
FindStringInTextWidget_( gchar*   pSought,
                         gboolean tryWrap,
                         gboolean quiet,
                         gboolean scroll,
                         gint*    pReturnLineNumFound,
                         gchar*   pText,
                         gint     textLen )
{
   gchar*	pMatch;
   gint		idxPos;

  /*
   * Compute the index of the char after the current position.
   */
   if ( GTK_EDITABLE(pWidgetText)->selection_end_pos )
      idxPos = GTK_EDITABLE(pWidgetText)->selection_end_pos;
   else
      idxPos = GTK_EDITABLE(pWidgetText)->current_pos;
   g_return_val_if_fail( idxPos <= textLen, FALSE );

  /*
   * Begin the search in the segment from the cursor,
   * direction depending on "findReverse".
   *
   * Skip this search if the cursor is at the start or end,
   * depending on search direction.
   */
   if (    (idxPos < textLen  &&  ! findReverse )
        || (idxPos >       0  &&    findReverse ) )
   {
      if ( ! findReverse )
         pMatch = FindStringInSegment( pSought,
                                       pText + idxPos,
                                       textLen - idxPos );
      else
         pMatch = FindStringInSegment( pSought,
                                       pText,
                                       idxPos );
   }
   else
   {
     /*
      * Cursor is at top or bottom, so a wrap search is needed.
      */
      pMatch = NULL;
   }

   if ( pMatch )
   {
      g_return_val_if_fail( pMatch >= pText, FALSE );

     /*
      * Found the string.
      * If requested, scroll/highlight string.
      */
      MostlyClearStatusbar();
      if ( scroll )
      {
         HighlightTextString( pText,
                              pMatch - pText,
                              strlen( pSought ) );
      }
      if ( pReturnLineNumFound )
      {
         *pReturnLineNumFound = LineIdx2LineNum( pText,
                                                 pMatch - pText );
      }

     /*
      * Done.
      */
      return TRUE;
   }

  /*
   * Not found.  Try a wrapped search (if specified to do so).
   * Is the cursor already at the top/bottom such that a full search was done?
   * If so, skip this wrapped search.
   */
   if ( ! tryWrap ) return FALSE;

   if ( idxPos > 0  &&  ! findReverse )
   {
     /*
      * Forward wrapped search.
      */
      pMatch = FindStringInSegment( pSought,
                                    pText,
                                    idxPos );
   }
   else if ( idxPos < textLen  &&  findReverse )
   {
     /*
      * Reverse wrapped search.
      * Ensure parms to pass will be correct.
      */
g_assert( textLen - idxPos >= 0 );
      if ( textLen - idxPos >= 0 )
         pMatch = FindStringInSegment( pSought,
                                       pText + idxPos,
                                       textLen - idxPos );
      else
         pMatch = NULL;
   }
   else
   {
      pMatch = NULL;
   }
   if ( pMatch )
   {
      g_return_val_if_fail( pMatch >= pText, FALSE );

     /*
      * Found the string.
      * Scroll to, and highlight, the string.
      */
      if ( ! quiet ) PrintStatusbarThirds( "Search for string '", pSought, "' wrapped." );
      HighlightTextString( pText,
                           pMatch - pText,
                           strlen( pSought ) );

     /*
      * Done.
      */
      return TRUE;
   }

  /*
   * String was definitely not found.
   */
   if ( ! quiet ) PrintStatusbarThirds( "String '", pSought, "' not found." );
   return FALSE;
}

/******************************************************************************
 * Main function for Find/Global that:
 *
 * - Finds all instances of a string in all modules.
 * - Builds a glist of TAG structs defining lines which contain sought string.
 * - Creates a notebook page containing a list widget, so that the user
 *   can select any match.
 *
 * Parms   : pSought
 *           Pointer to the string to find.
 *
 *           quiet
 *           If true then don't print to statusbar.
 *
 *           pGlistFoundTags (OUT/GLOBAL)
 *           List of tagged lines where matches were found.
 *
 *           findCaseSensText (GLOBAL)
 ******************************************************************************/
gboolean
FindStringGlobal( gchar*   pSought,
                  gboolean quiet,
                  gboolean stopIfFindDialogClosed )
{
   int res;
   gboolean FindStringGlobal___( gchar*   pSought,
                                 gboolean quiet,
                                 gboolean stopIfFindDialogClosed );

  /*
   * For stability, serialize this function.
   */
   if ( busy_FindStringGlobal ) { Warning( "FindStringGlobal() is busy." ); return FALSE; }
   busy_FindStringGlobal = TRUE;
   res = FindStringGlobal___( pSought, quiet, stopIfFindDialogClosed );
   busy_FindStringGlobal = FALSE;
   return res;
}
gboolean
FindStringGlobal___( gchar*   pSought,
                     gboolean quiet,
                     gboolean stopIfFindDialogClosed )
{
   gboolean notCanceled;

   PrintStatusbarNoTimeout( "Searching globally (please wait)..." );

  /*
   * Search all modules for all instances of sought string.
   * Output will be in pGlistFoundTags.
   */
   notCanceled = \
   BuildListOfFoundStrings( pSought,
                            quiet,
                            stopIfFindDialogClosed );

  /*
   * Were any matches found?
   */
   if ( pGlistFoundTags == NULL )
   {
      PrintStatusbar( "Nothing found (global search)." );

    /* TODO remove "found" notebook page, or clear found clist, or do nothing??? */

      return FALSE;
   }

  /*
   * Matches were found.
   * Create a notebook page showing the results of the search.
   */
   BuildNotebookPageWithFindResults();

   if ( notCanceled) MostlyClearStatusbar();
   FindDialogClose();

   return TRUE;
}

/******************************************************************************
 * Find all instances of a string in all modules.
 * Build a glist of TAG structs defining lines which contain sought string.
 * If discardTextBuffers then all module buffers will be freed.
 *
 * Parms   : pSought
 *           Pointer to the string to find.
 *
 *           quiet
 *           If true then don't print to statusbar.
 *
 *           pGlistFoundTags (OUT/GLOBAL)
 *           List of tagged lines where matches were found.
 *
 *           findCaseSensText (GLOBAL)
 *
 * Returns : TRUE if search completed (wasn't canceled/aborted).
 ******************************************************************************/
gboolean
BuildListOfFoundStrings( gchar*   pSought,
                         gboolean quiet,
                         gboolean stopIfFindDialogClosed )
{
   GList*	pModuleName			= NULL;
   gchar*	pModuleBuffer		= NULL;
   long		moduleSize;
   gchar*   pFound              = NULL;
   gchar*   pTextRemaining		= NULL;
   int      modules             = 0;
   gint     loops				= 0;
   gint     percentageShown     = 0;
   gint     percent;

   modules = g_hash_table_size( pHashModules );

  /*
   * Free previous glist.
   */
   if ( pGlistFoundTags ) g_list_free( pGlistFoundTags );
   pGlistFoundTags = NULL;

  /*
   * Search thru all modules.
   * Iterate for each name in the list "pGlistModuleNames".
   */
   for ( pModuleName = pGlistModuleNames;
         pModuleName;
         pModuleName = g_list_next( pModuleName ) )
   {
      ++loops;

      if ( (loops & 0x3f) == 0 )
      {
         percent = (int) ( (float)loops / (float)modules * 100 );
         if ( (percent - percentageShown) >= 5 )
         {
            percentageShown = percent;
            StatusbarShowPercentage( "Modules searched (please wait): __%.", percent );
         }
      }

      ProcessPendingEvents();

     /*
      * Stop if find dialog was closed (if requested so).
      */
      if ( stopIfFindDialogClosed  &&  pWidgetDialogFind == NULL )
      {
         PrintStatusbar( "Search canceled." );
         return FALSE;
      }

     /*
      * In order to search, load the next module into a memory buffer.
      */
      pModuleBuffer = (gchar*)ModuleBuffer( (gchar*)pModuleName->data,
                                            &moduleSize );
      if ( pModuleBuffer == NULL  ||  moduleSize == 0 )
         goto iterate; /* NULL can happen if file is empty */

     /*
      * Find all matches in this module.
      */
      pTextRemaining = pModuleBuffer;
      for (;;)
      {
        /*
         * Without this, a prolonged search will appear to have hanged the program.
         */
         static int loopThisModule = 0;
         if ( ++loopThisModule & 0x1f ) ProcessPendingEvents();

        /*
         * Search remaining text in buffer for the sought string.
         */
         pFound = FindStringInSegment( pSought,
                                       pTextRemaining,
                                       moduleSize - (pTextRemaining - pModuleBuffer) );

         #ifdef HYPERSRC_PEDANTIC
         if (    (loops > max(MAX_LOOPS, modules))
              || ( pFound != NULL  &&  pFound < pModuleBuffer)
              || (moduleSize - (pTextRemaining - pModuleBuffer) < 0 )
              || (pTextRemaining > (pModuleBuffer + moduleSize)) )
         { Warning( "Good programmers are hard to find.  Aborting BuildListOfFoundStrings()" ); return FALSE; }
         #endif

        /*
         * Was it found?
         */
         if ( pFound )
         {
           /*
            * Found, create a tag for this line, append tag to list.
            */
            AppendFoundTagsList( pSought,
                                 (gchar*)pModuleName->data,
                                 1 + LineIdx2LineNum( pModuleBuffer, pFound - pModuleBuffer ) );

           /*
            * Point after this match. 
            * Break if end of buffer.
            */
            pTextRemaining = pFound + strlen( pSought );
            if ( (pTextRemaining - pModuleBuffer) >= moduleSize ) break;

           /*
            * Continue searching for more matches in this module.
            */
            continue;
         }
         else
         {
           /*
            * Not found in this module, break to next module.
            */
            break;
         }
      }

     /*
      * Iterate to next module.
      */
      iterate:
      if ( discardTextBuffers ) FreeModuleBuffer( (gchar*)pModuleName->data );
   }

   if ( discardTextBuffers && pModuleName != NULL )
      FreeModuleBuffer( (gchar*)pModuleName->data );

  /*
   * Done.
   */
   MostlyClearStatusbar();
   return TRUE;
}

/******************************************************************************
 * Create a TAG struct for a line with a matching string.
 * Append the TAG struct to the global glist "pGlistFoundTags"
 * (this glist will be created if empty).
 *
 * Parms   : pTagName
 *           Written to TAG.pName.
 *
 *           pModuleName
 *           Written to TAG.pModule.
 *
 *           lineNum
 *           Written to TAG.pLine and TAG.lineInt
 *
 *           pGlistFoundTags (OUT/GLOBAL)
 ******************************************************************************/
FUNC_X
AppendFoundTagsList( gchar* pTagName,
                     gchar* pModuleName,
                     gint   lineNum )
{
   struct TAG*	pTag;

  /*
   * Allocate a TAG struct.
   */
   pTag = AllocTagStruct();

  /*
   * Fill the TAG struct.
   */
   FillTagStruct( pTag,
                  pTagName,
                  lineNum,
                  "found",
                  pModuleName );

  /*
   * Add the TAG struct to the list.
   */
   pGlistFoundTags = g_list_append( pGlistFoundTags, (gpointer)pTag );
   g_return_val_if_fail( pGlistFoundTags, FUNC_FAIL );

   return FUNC_OK;
}

/******************************************************************************
 * Create a notebook page with the results of a search.
 *
 * Parms   : pGlistFoundTags (GLOBAL)
 *           List that defines the search results.
 ******************************************************************************/
void
BuildNotebookPageWithFindResults( void )
{
   GtkWidget* pWidgetLabel;

  /*
   * Create a clist widget to hold each tagged matching line (unless it already exists).
   */
   if ( pWidgetClistFound == NULL )
   {
      CreateInvisibleTagsClistWidget( &pWidgetScrolledFound,
                                      &pWidgetClistFound,
                                      TAG_COL_LAYOUT_TAG_MODULE_LINE,
                                      (PFUNC)FoundClistWidgetRowSelectCallback,
                                      (PFUNC)FoundClistWidgetRowUnselectCallback );
      gtk_widget_show( pWidgetScrolledFound );
      gtk_widget_show( pWidgetClistFound );
   }
   else
   {
     /*
      * Clist widget already exists, remove all of its rows.
      */
      gtk_clist_clear( GTK_CLIST(pWidgetClistFound) );
   }

  /*
   * Add the tags clist widget (contained by a scrolled widget) to notebook.
   */
   if ( ! notebookPageFound )
   {
      pWidgetLabel = gtk_label_new( "found" );
      gtk_notebook_append_page( GTK_NOTEBOOK(pWidgetNotebook),
                                pWidgetScrolledFound,
                                pWidgetLabel );
      notebookPageFound = TRUE;
   }

  /*
   * Populate the clist widget with the search results.
   */
   gtk_clist_freeze( GTK_CLIST(pWidgetClistFound) );
   g_list_foreach( pGlistFoundTags,
                   AddRowToFoundClistWidget,
                   NULL );
   gtk_clist_thaw( GTK_CLIST(pWidgetClistFound) );

  /*
   * Select the notebook page.
   */
   gtk_notebook_set_page( GTK_NOTEBOOK(pWidgetNotebook),
                          NotebookPageFound() );

  /*
   * Install handlers when notebook page gains/loses focus.
   */
   InstallFocusHandlers( pWidgetClistFound );

   return;
}

/******************************************************************************
 * Using an element in the clist of tagged search results, add a row
 * to the clist widget containing search results.
 *
 * Each row is associated with a TAG struct (via the GTK "row data" mechanism).
 *
 * Parms   : pData
 *           Pointer to a TAG struct.
 *
 *         : pUnused
 *           (not used)
 ******************************************************************************/
void
AddRowToFoundClistWidget( gpointer pData,
                          gpointer pUnused )
{
   struct TAG*	pTag	= pData;
   gchar*       szRow[WIDGET_FOUND_CLIST_COLUMNS];
   int			row;

  /*
   * Actually add a row to the clist widget.
   */
   szRow[0] = pTag->pName;
   szRow[1] = pTag->pModule;
   szRow[2] = pTag->pLine;
   row = gtk_clist_append( GTK_CLIST(pWidgetClistFound),
                           szRow );

  /*
   * Associate the pointer to the TAG struct with the row.
   * The pointer will be needed later when the user selects a row.
   */
   gtk_clist_set_row_data( GTK_CLIST(pWidgetClistFound),
                           row,
                           pData );
   return;
}

/******************************************************************************
 * Callback for when a row is selected in the search results clist widget.
 *
 * The purpose of this callback is to load the text widget
 * with the module associated with the selected tag.
 *
 * Will automatically select the corresponding row in the modules clist,
 * thus indirectly calling ModulesClistWidgetRowSelectCallback().
 *
 * Parms   : pWidget
 *           Pointer to clist widget.
 *
 *         : row
 *           Which row was selected.
 *
 *         : col
 *           Which column was selected.
 *
 *         : GdkEventButton
 *           (not used here?)
 *
 *         : pData
 *           (not used here, NULL is passed)
 ******************************************************************************/
void
FoundClistWidgetRowSelectCallback( GtkWidget*      pWidget,
                                   gint            row,
                                   gint            col,
                                   GdkEventButton* pEvent,
                                   gpointer	       pData )
{
   struct TAG*	pTag;

  /*
   * Prepare to load the module containing the tag, and move the cursor over the tag.
   *
   * Obtain the pointer to the TAG struct that is associated
   * with the selected row in the tags clist widget.
   */
   pTag = (struct TAG*)gtk_clist_get_row_data( GTK_CLIST(pWidget), row );
g_return_if_fail( pTag );

  /*
   * Load the selected module (ie source file) into the text widget.
   * Move the cursor to the line that is tagged.
   */
   if ( LoadTextWidgetWithModule( pTag->pModule,
                                  pTag->lineInt - 1,
                                  row,
                                  DONT_RELOAD )
        != FUNC_OK )
   {
     /*
      * Oops, unable to load the module.
      * (A warning has already been displayed.)
      */
      return;
   }

  /*
   * Highlight the tag in the text widget.
   */
   HighlightWordInCurrentLine( pTag->pName );

  /*
   * Select the corresponding row in the modules clist.
   * Set a flag to inform the module's select_row callback to return early,
   * without affecting other colateral state such as the previous location etc.
   */
   returnEarlyModuleSelectCallback = TRUE;
   SearchClistWidget( pTag->pModule,
                      pWidgetClistModules,
                      0,
                      TRUE, /* exact match */
                      findCaseSensModule,
                      TRUE, /* be quiet */
                      FALSE,/* don't affect notebook */
                      TRUE	/* do select row */ );
   returnEarlyModuleSelectCallback = FALSE;

   return;
}

/******************************************************************************
 * Callback for when a row is UNselected in the search results clist widget.
 ******************************************************************************/
void
FoundClistWidgetRowUnselectCallback( GtkWidget*      pWidget,
                                     gint            row,
                                     gint            col,
                                     GdkEventButton* pEvent,
                                     gpointer	     pData )
{
  /*
   * (Yep, is a NOP.)
   */
   return;
}
