/******************************************************************************
 * $Id: hilite.c,v 1.23 2000/12/31 18:05:42 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for hiliting text in a text widget.
 * Notes        : Because of the huge function HiliteStringOfCCode(),
 *                this code is very slow.  Slowness becomes a problem
 *                while trying to hilite a source file > 1MB because
 *                the program will appear to hang.
 ******************************************************************************/

#include "header.h"

/*
 * Color used for hilites.
 */
GdkColor colorComment  = { 0, 0x0800, 0x7000, 0x0800 };
GdkColor colorKeyword  = { 0, 0x1000, 0x1000, 0xb000 };
GdkColor colorType     = { 0, 0xb000, 0x1000, 0xc000 };
GdkColor colorSymbol   = { 0, 0xb000, 0x1000, 0x1000 };


/*****************************************************************************
 * Load a text widget with highlighted C source code.
 *
 * Precond : The caller must first empty and freeze the text widget.
 *
 * Parms   : pWidget
 *           Pointer to a text widget.
 *
 *           pBuffer
 *           Pointer to a buffer containing source code text.
 *           Buffer must have a trailing null char.
 *
 *           bufferSize
 *           Size of buffer (excluding trailing null char).
 *****************************************************************************/
void
HiliteLoadCCode( GtkWidget* pWidget,
                 char*      pBuffer,
                 int        bufferSize )
{
   const char*	pc;
   const char*	pcAfter;
   const char*	pcTmp1;
   const char*	pcTmp2;
   const char*	pcPrePastWhitespace;
   const char*	pCommentAfter;
   GdkFont*		pFontComment		= NULL;
   GdkFont*		pFontKeyword		= NULL;
   GdkFont*		pFontType			= NULL;
   GdkFont*		pFontSymbol			= NULL;
   char*		pScratch;
   gboolean		useHiliteFontsOutsideComments = TRUE;
   gboolean		useHiliteFontsInsideComments  = TRUE;
   int			textWidgetSize;

#define REMAINING_BUFFER_SIZE(p)	(bufferSize - (p - pBuffer))

  /*
   * Ensure the buffer is terminated by a null char,
   * otherwise the parsing of it would overrun.
   */
   g_assert( pBuffer[bufferSize] == '\0' );

  /*
   * Is buffer tiny or too big?
   * This code is so slow that trying to hilite a source file > 1MB
   * will make the program appear to hang.
   */
   if ( bufferSize < 10  ||  bufferSize > MAX_HILITE_BUFFER_LEN )
   {
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pBuffer,
                       bufferSize );

      if ( bufferSize > MAX_HILITE_BUFFER_LEN )
         PrintStatusbar( "Buffer is too large to highlight." );

      return;
   }

  /*
   * Prepare fonts.
   */
   if ( useHiliteFontsInsideComments )
   {
      pFontComment = gdk_font_load( fontNameHiliteComment );
   }
   if ( useHiliteFontsOutsideComments )
   {
      pFontKeyword = gdk_font_load( fontNameHiliteKeyword );
      pFontType    = gdk_font_load( fontNameHiliteType );
      pFontSymbol  = gdk_font_load( fontNameHiliteSymbol );
   }

  /*
   * Allocate mem for a temp string as large as module buffer.
   * Will be re-used many times by lower-level functions.
   */
   pScratch = g_malloc( bufferSize );

  /*
   * Parse buffer for strings to hilite.
   */
   for ( pc = pBuffer; pc < &pBuffer[bufferSize]; )
   {
     /*
      * Note: Loop should be stopped using break, not return.
      */
      g_assert( pc >= pBuffer  &&  pc < &pBuffer[bufferSize] );

     /*
      * Has end of buffer been reached?
      */
      g_assert( pc != NULL );	/* should not be NULL at this point */
      if ( pc >= &pc[bufferSize]  ||  *pc == '\0' )
         break;

     /*
      * Move past whitespace, but remember prior pointer.
      */
      pcPrePastWhitespace = pc;
      pc = (char*)PastWhitespace( pc, REMAINING_BUFFER_SIZE(pc) );

     /*
      * Does only whitespace remain?
      */
      if ( pc == NULL )
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          NULL,
                          NULL,
                          NULL,
                          pcPrePastWhitespace,
                          REMAINING_BUFFER_SIZE(pcPrePastWhitespace ) );
         break;
      }

     /*
      * Is the beginning of a C/C++ comment being pointed to?
      */
      if (    (pc[0] == '/'  &&  pc[1] == '*') 	/* trailing \0 obviates checking ptr near buffer end */
           || (pc[0] == '/'  &&  pc[1] == '/') )
      {
        /*
         * Parse/insert C comment.
         */
         pCommentAfter = \
         HiliteParseInsertComment( pWidget,
                                   &colorComment,
                                   pFontComment,
                                   pcPrePastWhitespace,
                                   (pc[1] == '*') ? cCommentEndChars : "\n",
                                   REMAINING_BUFFER_SIZE(pcPrePastWhitespace) );

        /*
         * Was the end of the C comment parsed?
         */
         if ( pCommentAfter )
         {
           /*
            * From after C comment, iterate.
            */
            pc = pCommentAfter;
            continue;
         }
         else
         {
           /*
            * End of buffer must have been prematurely encountered.
            */
            break;
         }
      }

     /*
      * Try to scoop contiguous symbol chars.
      */
      if ( ! isalpha( *pc )  )	/* perhaps non-alpha is a symbol char? */
      {
        /*
         * For each definite symbol char after any whitespace...
         */
         for( pcAfter = pc;
              IsCSymbolChar( *pcAfter );	/* trailing \0 obviates checking ptr near end */
              ++pcAfter )
#ifndef HYPERSRC_PEDANTIC
         { ; }
#else
         { g_assert( pcAfter < &pBuffer[bufferSize] ); }
#endif

        /*
         * Were any symbol chars counted?
         */
         if ( pcAfter > pc )
         {
            gtk_text_insert( GTK_TEXT(pWidget),
                             pFontSymbol,
                             &colorSymbol,
                             NULL,
                             pcPrePastWhitespace,
                             pcAfter - pcPrePastWhitespace );
           /*
            * Move pointer, continue (stopping cases are checked above).
            */
            pc = pcAfter;
            continue;
         }

        /*
         * Fall thru.
         */
      }

     /*
      * Scoop all text until: start of C comment, or symbol char, or buffer end.
      */
      pcTmp1 = strstr( pc, cCommentStartChars );
      pcTmp2 = FindCSymbolChar( pc, REMAINING_BUFFER_SIZE(pc) );

      if ( pcTmp1 == NULL )					/* process lower ptr unless it is NULL */
         pcAfter = pcTmp2;
      else if ( pcTmp2 == NULL )
         pcAfter = pcTmp1;
      else
         pcAfter = min( pcTmp1, pcTmp2 );	/* neither was NULL */

      if ( pcAfter == NULL )				/* was both NULL (buffer end)? */
      {
        /*
         * Encountered buffer end.
         */
         pcAfter = &pBuffer[bufferSize];
      }

#if 0
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pcPrePastWhitespace,
                       pcAfter - pcPrePastWhitespace );
#else
      HiliteStringOfCCode( pWidgetText,
                           pFontKeyword,
                           pFontType,
                           pFontSymbol,
                           &colorKeyword,
                           &colorType,
                           &colorSymbol,
                           pcPrePastWhitespace,
                           pcAfter - pcPrePastWhitespace,
                           pScratch );
#endif

     /*
      * Move pointer, continue (stopping cases are checked above).
      */
      pc = pcAfter;
   }

  /*
   * There's a bug here if the size of the text widget
   * does not equal the size of the buffer.
   */
   textWidgetSize = gtk_text_get_length( GTK_TEXT(pWidgetText) );
   if ( textWidgetSize != bufferSize )
   {
      Warning( "HiliteLoadCCode(): gtk_text_get_length() != bufferSize in module %s", moduleLoadedIntoTextWidget );
      Warning( "HiliteLoadCCode(): gtk_text_get_length() = %d  bufferSize = %d", textWidgetSize, bufferSize );
   }

  /*
   * Dealloc, return.
   */
   g_free( pScratch );
   return;
}

/*****************************************************************************
 * Parse the comment that is being pointed to.
 * Insert the comment as hilited text into a text widget.
 *
 * This function is generalized.  It can accomodate C/C++ comments,
 * perhaps other langs.
 *
 * Parms   : pWidget
 *           Pointer to a text widget.
 *
 *           pFont
 *           Hilite font (NULL can be passed).
 *
 *           pColor
 *           Hilite color (NULL can be passed).
 *
 *           pCommentStartApprox
 *           This can point to either the exact start of a C comment (to "/"),
 *           or can point to whitespace that precedes the C comment.
 *
 *			 pCommentEndDef
 *           Pointer to a string that defines what the end of comment is.
 *           Eg, "/""*" or "\n" for C or C++, resp.
 *
 *           bufferSize
 *           Size of buffer, measurement based on "pCommentStart" as the start.
 *
 * Returns : If a C comment end was present, returns a pointer after it,
 *           or returns NULL.
 *****************************************************************************/
const char*
HiliteParseInsertComment( GtkWidget*  pWidget,
                          GdkColor*   pColor,
                          GdkFont*    pFont,
                          const char* pCommentStartApprox,
                          const char* pCommentEndDef,
                          int         bufferSize )
{
   char*	pCommentEnd;
   int      commentLen;

#ifdef HYPERSRC_PEDANTIC
   g_assert( pCommentStartApprox  &&  pCommentStartApprox[0]  &&  bufferSize > 0 );
#endif

  /*
   * Try to find end of comment.
   */
   pCommentEnd = strstr( pCommentStartApprox, pCommentEndDef );

  /*
   * Compute length of C comment.
   * If the C comment epilog is missing then use end-of-buffer.
   */
   if ( pCommentEnd != NULL )
      commentLen = (pCommentEnd + strlen(pCommentEndDef)) - pCommentStartApprox;
   else
      commentLen = bufferSize;

  /*
   * Insert C comment into text widget.
   */
   gtk_text_insert( GTK_TEXT(pWidget),
                    pFont,
                    pColor,
                    NULL,
                    pCommentStartApprox,
                    commentLen );

  /*
   * Return pointer to after end of comment, or NULL.
   */
   if ( pCommentEnd )
      return pCommentEnd + strlen(pCommentEndDef);
   else
      return NULL;
}

/*****************************************************************************
 * Parse a string of C/C++ code (ie text outside of comment enclosures).
 *
 * Parms   : pWidget
 *           Pointer to a text widget.
 *
 *           pFont[Keyword,Type,Symbol]
 *           Hilite font (NULL can be passed).
 *
 *           pColor[Keyword,Type,Symbol]
 *           Hilite color (NULL can be passed).
 *
 *           pStr
 *			 Pointer to a string of C code (uncommented).
 *
 *           strSize
 *           Size of string of C code.
 *
 *           pTmp
 *           Should point to a buffer as large as corresponding modue buffer.
 *
 * Returns : If a C comment end was present, returns a pointer after it,
 *           or returns NULL.
 *****************************************************************************/
void
HiliteStringOfCCode( GtkWidget*  pWidget,
                     GdkFont*	 pFontKeyword,
                     GdkFont*	 pFontType,
                     GdkFont*	 pFontSymbol,
                     GdkColor*	 pColorKeyword,
                     GdkColor*	 pColorType,
                     GdkColor*	 pColorSymbol,
                     const char* pStr,
                     int         strSize,
                     char*		 pScratch )
{
   const char*	pc = NULL;
   const char*	pWord;
   int          sizeWord;

   if ( strSize == 0 ) return;
   g_assert( pStr && strSize > 0 );

   for ( pc = pStr;
         pc < &pStr[strSize];
         pc = (char*)PastWord( pWord, sizeWord ) )
   {
     /*
      * Parse next word.
      */
      sizeWord = ParseWordAlphanum( pc,
                                    &pWord,
                                    strSize - (pc - pStr) );
     /*
      * If no word was parsed (or word was too big to parse),
      * insert remainder into text widget, return.
      */
      if ( sizeWord <= 0  ||  pWord == NULL )	/* ParseWord() can return negative */
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          NULL,
                          NULL,
                          NULL,
                          pc,
                          strSize - (pc - pStr) );
         return;
      }

#ifdef HYPERSRC_PEDANTIC
g_assert( pWord >= pc  &&  pWord < &pStr[strSize] );
#endif

     /*
      * For comparing strings, extract word into a ASCIIZ string.
      */
      memcpy( pScratch, pWord, sizeWord );
      pScratch[sizeWord] = '\0';

     /*
      * Is word a C/C++ keyword for non-types?
      */
      if ( /* C keywords */
              strcmp( pScratch, "if"        ) == 0
           || strcmp( pScratch, "else"      ) == 0
           || strcmp( pScratch, "return"    ) == 0
           || strcmp( pScratch, "goto"      ) == 0
           || strcmp( pScratch, "for"       ) == 0
           || strcmp( pScratch, "while"     ) == 0
           || strcmp( pScratch, "do"        ) == 0
           || strcmp( pScratch, "continue"  ) == 0
           || strcmp( pScratch, "break"     ) == 0
           || strcmp( pScratch, "switch"    ) == 0
           || strcmp( pScratch, "case"      ) == 0
           || strcmp( pScratch, "default"   ) == 0
           || strcmp( pScratch, "sizeof"    ) == 0
           || strcmp( pScratch, "asm"       ) == 0

           /* preprocessor keywords */
           || strcmp( pScratch, "#define"   ) == 0
           || strcmp( pScratch, "#if"       ) == 0
           || strcmp( pScratch, "#else"     ) == 0
           || strcmp( pScratch, "#elif"     ) == 0
           || strcmp( pScratch, "#endif"    ) == 0
           || strcmp( pScratch, "#ifdef"    ) == 0
           || strcmp( pScratch, "#ifndef"   ) == 0
           || strcmp( pScratch, "#include"  ) == 0
           || strcmp( pScratch, "#undef"    ) == 0
           || strcmp( pScratch, "#error"    ) == 0
           || strcmp( pScratch, "#pragma"   ) == 0

           /* C++ keywords */
           || strcmp( pScratch, "catch"     ) == 0
           || strcmp( pScratch, "delete"    ) == 0
           || strcmp( pScratch, "inline"    ) == 0
           || strcmp( pScratch, "new"       ) == 0
           || strcmp( pScratch, "operator"  ) == 0
           || strcmp( pScratch, "this"      ) == 0
           || strcmp( pScratch, "throw"     ) == 0
           || strcmp( pScratch, "try"       ) == 0 )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontKeyword,
                          pColorKeyword,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is word a C/C++ keyword for types?
      */
      if (    strcmp( pScratch, "void"      ) == 0
           || strcmp( pScratch, "char"      ) == 0
           || strcmp( pScratch, "short"     ) == 0
           || strcmp( pScratch, "int"       ) == 0
           || strcmp( pScratch, "long"      ) == 0
           || strcmp( pScratch, "float"     ) == 0
           || strcmp( pScratch, "double"    ) == 0
           || strcmp( pScratch, "const"     ) == 0
           || strcmp( pScratch, "volatile"  ) == 0
           || strcmp( pScratch, "register"  ) == 0
           || strcmp( pScratch, "static"    ) == 0
           || strcmp( pScratch, "extern"    ) == 0
           || strcmp( pScratch, "signed"    ) == 0
           || strcmp( pScratch, "unsigned"  ) == 0
           || strcmp( pScratch, "struct"    ) == 0
           || strcmp( pScratch, "union"     ) == 0
           || strcmp( pScratch, "typedef"   ) == 0
           || strcmp( pScratch, "enum"      ) == 0

           /* C++ */
           || strcmp( pScratch, "bool"      ) == 0
           || strcmp( pScratch, "class"     ) == 0
           || strcmp( pScratch, "friend"    ) == 0
           || strcmp( pScratch, "private"   ) == 0
           || strcmp( pScratch, "protected" ) == 0
           || strcmp( pScratch, "public"    ) == 0
           || strcmp( pScratch, "template"  ) == 0
           || strcmp( pScratch, "virtual"   ) == 0

           /* non-standard C */
           || strcmp( pScratch, "void*"     ) == 0
           || strcmp( pScratch, "char*"     ) == 0
           || strcmp( pScratch, "short*"    ) == 0
           || strcmp( pScratch, "int*"      ) == 0
           || strcmp( pScratch, "long*"     ) == 0

           || strcmp( pScratch, "byte*"     ) == 0
           || strcmp( pScratch, "byte*"     ) == 0
           || strcmp( pScratch, "uchar"     ) == 0
           || strcmp( pScratch, "uchar*"    ) == 0
           || strcmp( pScratch, "uint"      ) == 0
           || strcmp( pScratch, "uint*"     ) == 0
           || strcmp( pScratch, "int8"      ) == 0
           || strcmp( pScratch, "int8*"     ) == 0
           || strcmp( pScratch, "int16"     ) == 0
           || strcmp( pScratch, "int16*"    ) == 0
           || strcmp( pScratch, "int32"     ) == 0
           || strcmp( pScratch, "int32*"    ) == 0

           /* GLIB */
           || strcmp( pScratch, "gpointer"  ) == 0
           || strcmp( pScratch, "gconstpointer") == 0
           || strcmp( pScratch, "gboolean"  ) == 0
           || strcmp( pScratch, "gchar"     ) == 0
           || strcmp( pScratch, "gchar*"    ) == 0
           || strcmp( pScratch, "guchar"    ) == 0
           || strcmp( pScratch, "guchar*"   ) == 0
           || strcmp( pScratch, "gint"      ) == 0
           || strcmp( pScratch, "gint*"     ) == 0
           || strcmp( pScratch, "guint"     ) == 0
           || strcmp( pScratch, "guint*"    ) == 0
           || strcmp( pScratch, "glong"     ) == 0
           || strcmp( pScratch, "glong*"    ) == 0
           || strcmp( pScratch, "gulong"    ) == 0
           || strcmp( pScratch, "gulong*"   ) == 0
           || strcmp( pScratch, "gfloat"    ) == 0
           || strcmp( pScratch, "gfloat*"   ) == 0
           || strcmp( pScratch, "gint8"     ) == 0
           || strcmp( pScratch, "gint8*"    ) == 0
           || strcmp( pScratch, "gint16"    ) == 0
           || strcmp( pScratch, "gint16*"   ) == 0
           || strcmp( pScratch, "gint32"    ) == 0
           || strcmp( pScratch, "gint32*"   ) == 0
           || strcmp( pScratch, "gint64"    ) == 0
           || strcmp( pScratch, "gint64*"   ) == 0 )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontType,
                          pColorType,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is word a C/C++ symbol char?
      */
      if (    strcmp( pScratch, "="         ) == 0
           || strcmp( pScratch, "=="        ) == 0
           || strcmp( pScratch, "!="        ) == 0
           || strcmp( pScratch, "<"         ) == 0
           || strcmp( pScratch, ">"         ) == 0
           || strcmp( pScratch, "<="        ) == 0
           || strcmp( pScratch, ">="        ) == 0
           || strcmp( pScratch, "+"         ) == 0
           || strcmp( pScratch, "-"         ) == 0
           || strcmp( pScratch, "*"         ) == 0
           || strcmp( pScratch, "/"         ) == 0
           || strcmp( pScratch, "%"         ) == 0
           || strcmp( pScratch, "||"        ) == 0
           || strcmp( pScratch, "&&"        ) == 0
           || strcmp( pScratch, "?"         ) == 0
           || strcmp( pScratch, "!"         ) == 0
           || strcmp( pScratch, "&"         ) == 0
           || strcmp( pScratch, "|"         ) == 0
           || strcmp( pScratch, "^"         ) == 0
           || strcmp( pScratch, "~"         ) == 0
           || strcmp( pScratch, ","         ) == 0
           || strcmp( pScratch, "["         ) == 0
           || strcmp( pScratch, "]"         ) == 0
           || strcmp( pScratch, "."         ) == 0
           || strcmp( pScratch, "<<"        ) == 0
           || strcmp( pScratch, ">>"        ) == 0
           || strcmp( pScratch, "++"        ) == 0
           || strcmp( pScratch, "--"        ) == 0
           || strcmp( pScratch, "+="        ) == 0
           || strcmp( pScratch, "-="        ) == 0
           || strcmp( pScratch, "*="        ) == 0
           || strcmp( pScratch, "/="        ) == 0
           || strcmp( pScratch, "%="        ) == 0
           || strcmp( pScratch, "&="        ) == 0
           || strcmp( pScratch, "^="        ) == 0
           || strcmp( pScratch, "|="        ) == 0
           || strcmp( pScratch, "<<="       ) == 0
           || strcmp( pScratch, ">>="       ) == 0
           || strcmp( pScratch, ":"         ) == 0 )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontSymbol,
                          pColorSymbol,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }


     /*
      * Word was not recognized.
      * Using normal font/color, insert the parsed word
      * along with preceding whitespace.
      */
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pc,
                       (pWord - pc) + sizeWord );

     /*
      * Iterate...
      */
   }
}
