/******************************************************************************
 * $Id: module.c,v 1.17 2000/10/19 01:57:11 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for managing source code modules (ie files).
 ******************************************************************************/

#include "header.h"

/******************************************************************************
 * Return a pointer to a buffer that holds a source code module (ie file).
 *
 * The buffer has a null char appended that is NOT accounted
 * by the returned module size.
 *
 * If the module has not been loaded, then this will load it.
 *
 * NO NEED TO FREE BUFFER MEMORY DURING LIFE OF PROGRAM.
 *
 * Returns : Pointer to memory buffer allocated by ** g_malloc() **
 *           -or- NULL upon failure.
 *
 * Parms   : pModuleName
 *           Name of module.
 *
 *         : pModuleSize (out)
 *           Size of module (undefined if module could not be loaded).
 *           This does NOT account a null char that was appended to buffer.
 ******************************************************************************/
gpointer
ModuleBuffer( gchar* pModuleName,
              long*  pModuleSize /* out */ )
{
   P_HASH_ITEM_MODULE	pHashItem;

  /*
   * If the sought module has been loaded into a buffer,
   * then it will have a hash item that has a non-NULL pointer to the buffer.
   */
   pHashItem = (P_HASH_ITEM_MODULE)g_hash_table_lookup( pHashModules,
                                                        (gconstpointer)pModuleName );
   if ( pHashItem  &&  pHashItem->pModuleBuffer )
   {
     /*
      * Output the pointer to the buffer, and the size of the buffer.
      */
      g_return_val_if_fail( pHashItem->pModuleBuffer, NULL );
#if 0
     /* A buffer can legitimately be empty */
      g_assert( pHashItem->moduleSize );
#endif
      *pModuleSize = pHashItem->moduleSize;
      return pHashItem->pModuleBuffer;
   }

  /*
   * The module has never been loaded, so load it now (from a file).
   * Begin by creating a hash item struct.
   */
   pHashItem = (P_HASH_ITEM_MODULE)g_malloc( sizeof(HASH_ITEM_MODULE) );
   g_return_val_if_fail( pHashItem, NULL );
#ifdef HYPERSRC_PEDANTIC
   g_return_val_if_fail( sizeof(HASH_ITEM_MODULE), NULL );
#endif

  /*
   * Load the file into a buffer.
   * LoadFile() will append a null to buffer that is not
   * accounted in the returned size.
   * Consequently, fill the hash item struct.
   */
   pHashItem->pModuleBuffer = LoadFile( pModuleName,
                                        pModuleSize /* out */ );
   if ( pHashItem->pModuleBuffer == NULL )
   {
     /*
      * Oops, could not open the module.
      */
      PrintStatusbarThirds( "Could not open module '", pModuleName, "'." );
      Warning( "could not open module %s ", pModuleName );
      return NULL;
   }
   pHashItem->moduleSize = *pModuleSize;

#ifdef HYPERSRC_PEDANTIC
  /*
   * Check if a null char exists within the buffer,
   * because code which parses buffers would malfunction.
   * LoadFile() has appended the buffer with a null char.
   */
 {
   char* pNull = strchr( pHashItem->pModuleBuffer, '\0' );
   if ( pNull < (char*)pHashItem->pModuleBuffer + *pModuleSize )
   {
      Warning( "A null char found in middle of file %s \n", pModuleName );
      return NULL;
   }
 }
#endif

  /*
   * Insert/replace the hash item in the hash table.
   */
   g_hash_table_insert( pHashModules,
                        (gpointer)pModuleName,
                        (gpointer)pHashItem );

  /*
   * (The hash item struct will remain allocated.)
   */

  /*
   * Return a pointer to the buffer that holds the module.
   * (The size of the buffer was output to "pModuleSize".)
   */
   return pHashItem->pModuleBuffer;
}

/******************************************************************************
 * Free buffer that holds a module.
 * Updates corresponding module hash item.
 *
 * Parms   : pModuleName
 *           Name of module.
 *
 *         : pHashItem->pModuleBuffer (OUT/GLOBAL)
 *           pHashItem->moduleSize    (OUT/GLOBAL)
 *           Updated to indicate module was freed.
 ******************************************************************************/
void
FreeModuleBuffer( gchar* pModuleName )
{
   P_HASH_ITEM_MODULE	pHashItem;

  /*
   * Skip if name indicates if no module in text widget.
   */
   if (    pModuleName == NULL
        || pModuleName[0] == '\0' )
      return;

  /*
   * If the sought module has been loaded into a buffer,
   * then it will have a hash item that has a non-NULL pointer to the buffer.
   */
   pHashItem = (P_HASH_ITEM_MODULE)g_hash_table_lookup( pHashModules,
                                                        (gconstpointer)pModuleName );
   if ( pHashItem  &&  pHashItem->pModuleBuffer )
   {
     /*
      * Free the buffer.
      * Record in hash item that buffer was freed.
      */
      g_free( pHashItem->pModuleBuffer );
      pHashItem->pModuleBuffer = NULL;
      pHashItem->moduleSize    = 0;
   }

   return;
}

/******************************************************************************
 * Returns a hash value from a key.
 ******************************************************************************/
guint
HashModuleKey( gconstpointer pKey )
{
#if 1
   return g_str_hash( pKey );
#else
   static hashModuleKey = 0;

   return ++hashModuleKey;
#endif
}

/******************************************************************************
 * The comparsion function for the module hash.
 ******************************************************************************/
gint
HashModuleCmp( gconstpointer pName1,
               gconstpointer pName2 )
{
#if 1
   return g_str_equal( pName1,
                       pName2 );
#else
   return ! strcmp( (char*)pName1,
                    (char*)pName2 );
#endif
}
